pro zad13_2D, n_xx=n_x, n_t=n_t, sgma=sgma, frac=frac, rngs=rngs

  ; ustawienia (z argumentw nazwanych lub domylne)
  n_x  = keyword_set(n_x)  ? n_x  : 200             ; dugo wektora stanu
  n_t  = keyword_set(n_t)  ? n_t  : 400            ; liczba krokw czasowych
  sgma = keyword_set(sgma) ? sgma : 5.             ; odchylenie standardowe bdu obserwacji
  frac = keyword_set(frac) ? frac : .2             ; prawdopodobiestwo wykonania obserwacji
  rngs = keyword_set(rngs) ? rngs : long(10e7*!PI) ; ziarno dla generatora liczb pseudolosowych

  ; tablice z trzema realizacjami procesu markowa (prawdziwa, obserwowana i szacowana)
  X_tru = fltarr(n_x, n_t) 
  X_obs = replicate(!VALUES.F_NAN, n_x, n_t) 
  X_est = replicate(!VALUES.F_NAN, n_x, n_t)

  ; rozpatrujemy przykadowy proces Markowa: x(t) = F X(t-1) + w(t) 
  ; (ponisze linijki generuj jedynie macierz propagujc wymylny falowy kszta)
  F = fltarr(n_x, n_x)
  pnkt = indgen(n_x) - (n_x - 1.) / 2.
  wzor = exp(-(pnkt/20)^2 + complex(0,1)*2*!PI*pnkt/24)
  wzor_r = float(wzor)/sqrt(total(float(wzor)^2))
  wzor_i = imaginary(wzor)/sqrt(total(imaginary(wzor)^2))
  ksztalt = [[wzor_r], [wzor_i]]
  kat = !DTOR * 40
  obrot = [[cos(kat), -sin(kat)], [sin(kat), cos(kat)]]
  F = ksztalt # obrot # transpose(ksztalt)

  ; okrelenie macierzy kowariancji bdu (zadanie jej z gry)
  Q = fltarr(n_x, n_x)
  for j = 0, n_x - 1 do for i = 0, n_x - 1 do Q[i, j] = .9^(abs(i-j))

  ; rozpisujemy: w(t) na iloczyn macierzy i wektora wartoci losowych o rozkadzie norm.
  ; wyznaczamy E (za pomoc dekompozycji SVD), tak aby Q = E # transpose(E)
  svdc, Q, W, E, V, /column
  for j = 0, n_x - 1 do E[*, j] *= sqrt(W[j])

  ; ---------------------------------------------------------------------------------------------------
  ; ptla wyliczajca prawdziw realizacj oraz symulowane obserwacje
  for t = 0, n_t - 1 do begin
    ; ,,prawdziwa'' realizacja procesu (dla t=0 sam szum)
    X_tru[*, t] = t eq 0 ? 0 : F # X_tru[*, t - 1] + E # randomn(rngs, n_x) 

    ; ,,obserwowana'' realizacja procesu, czyli symulowane pomiary:
    ;  - obcione bdem o rozkadzie normalnym o odchyleniu standardowym _sgma_, 
    ;  - wykonywane dla czci _frac_ losowo wybranych punktw wektora stanu
    obsidx = where(randomu(rngs, n_x) lt frac, n_o)
    if n_o gt 0 then X_obs[obsidx, t] = X_tru[obsidx, t] + sgma * randomn(rngs, n_o)
  endfor

  ; ---------------------------------------------------------------------------------------------------
  ; Wyznaczenie realizacji ,,szacowanej'' za pomoc filtru Kalmana
  ; (P=Q bo znamy dokadnie macierz kowariancji bdu w modelu - zadalimy j powyej)

  ; znajduj t0 - numer pierwszego kroku w ktrym s pomiary, dla chwili X_est[t0] = mean(X_obs[t0]) 
  for t0 = 0, n_t - 1 do begin
    avg = mean(X_obs[*, t0], /nan)
    if finite(avg) then begin
      X_est[*, t0] = avg
      break
    endif
  endfor
  if t0 ge n_t - 1 then message, 'za mao krokw czasowych z pomiarami'

  for t = t0 + 1, n_t - 1 do begin

    ; odtworzenie obsidx z poprzedniej ptli (tylko dla pokzania, e to nie musi by w tej samej ptli)
    obsidx = where(finite(X_obs[*, t]), n_o)
    if n_o eq 0 then begin
      message, /continue, 'brak obserwacji w chwili' + strtrim(string(t), 2) + ', powielam X_est[t-1]'
      X_est[*, t] = X_est[*, t - 1]
      continue
    endif

    ; macierz kowariancji bdu obserwacji (sgma^2 na diagonali)
    R = fltarr(n_o, n_o)
    R[indgen(n_o) * (n_o + 1)] = sgma^2

    ; operator obserwacji (transformujcy wektor stanu do przestrzeni obserwacji)
    H = fltarr(n_o, n_x)
    if n_o gt 0 then for o = 0, n_o - 1 do H[o, obsidx[o]] = 1

    ; predykcja (nie zaley od obserwacji, zaley od modelu procesu)
    X_est[*, t] = F # X_est[*, t - 1]
    
    ; uaktualnienie (zaley od predykcji i obserwacji)
    K = Q # transpose(H) # invert((H # Q # transpose(H) + temporary(R)))
    X_est[*, t] += temporary(K) # (X_obs[obsidx, t] - temporary(H) # X_est[*, t])

  endfor

  ; ---------------------------------------------------------------------------------------------------
  ; wizualizacja wynikw

  loadct, 1
  zrange=max(abs(X_tru), /nan)
  window, 1, xsize=700, ysize=500
  tv, rebin(bytscl(X_tru, min=-zrange, max=zrange), 200, 400, /sample), 10, 10
  xyouts,  10, 440, 'prawdziwa', /device
  tv, rebin(bytscl(X_obs, min=-zrange, max=zrange), 200, 400, /sample), 250, 10
  xyouts, 250, 440,'obserwowana', /device
  tv, rebin(bytscl(X_est, min=-zrange, max=zrange), 200, 400, /sample), 490, 10
  xyouts, 490, 440, 'szacowana', /device

  window,2
  plot, X_tru(*, 50), line=1, title='Przyklad kroku czasowego (50)'
  oplot, X_est(*, 50)
  obsidx = where(finite(X_obs(*, 50)))
  oplot, obsidx, X_obs(obsidx, 50), psym=4

end
